#include "common.h"


typedef struct SLAVEINFO {
	LPSLAVEMSGDATA		lpMsg;
	DWORD				dwSize;
	HANDLE				hFileMap;
	HANDLE				hInEvent;
	HANDLE				hOutEvent;
	PROCESS_INFORMATION	piSlave;
} SLAVEINFO, *LPSLAVEINFO;


/** @return if FALSE, use GetLastError() */
BOOL CloseSlave(LPSLAVEINFO lpSlave, HRESULT* phr)
{
	BOOL	z	= TRUE;
	DWORD	err	= ERROR_SUCCESS;

	TRACE(trace(TEXT("*** [MASTER] CloseSlave($%08lX,$%08lX)\n"),lpSlave,phr));

	if(lpSlave->piSlave.hProcess) {
		TRACE(trace(TEXT("*** [MASTER] CloseSlave(): WaitForSingleObject()...\n")));
		if(WAIT_FAILED == WaitForSingleObject(lpSlave->piSlave.hProcess,INFINITE)) EXITONLASTERROR();
		if(phr) {
			TRACE(trace(TEXT("*** [MASTER] CloseSlave(): GetExitCodeProcess()...\n")));
			if(!GetExitCodeProcess(lpSlave->piSlave.hProcess,(DWORD*)phr)) EXITONLASTERROR();
			TRACE(trace(TEXT(">>> [MASTER] CloseSlave(): *phr = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(*phr)));
#ifdef AXL_TRACE
		} else {
			HRESULT hr;
			TRACE(trace(TEXT("*** [MASTER] CloseSlave(): GetExitCodeProcess()...\n")));
			if(!GetExitCodeProcess(lpSlave->piSlave.hProcess,(DWORD*)&hr)) EXITONLASTERROR();
			TRACE(trace(TEXT(">>> [MASTER] CloseSlave(): hr = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(hr)));
#endif
		}
	}

exit:
	if(lpSlave->piSlave.hThread) (void)CloseHandle(lpSlave->piSlave.hThread);
	if(lpSlave->piSlave.hProcess) (void)CloseHandle(lpSlave->piSlave.hProcess);
	if(lpSlave->lpMsg) (void)UnmapViewOfFile(lpSlave->lpMsg);
	if(lpSlave->hFileMap) (void)CloseHandle(lpSlave->hFileMap);
	if(lpSlave->hOutEvent) (void)CloseHandle(lpSlave->hOutEvent);
	if(lpSlave->hInEvent) (void)CloseHandle(lpSlave->hInEvent);
	(void)memset(lpSlave,0,sizeof(*lpSlave));
	TRACE(trace(TEXT("*** [MASTER] CloseSlave() >>> %s (%lu)\n"),z ? TEXT("TRUE") : TEXT("FALSE"),err));
	if(!z) SetLastError(err); return z;
}

/** @return if FALSE, use GetLastError() */
BOOL OpenSlave(LPSLAVEINFO lpSlave, DWORD dwInSize, DWORD dwOutSize)
{
	BOOL				z	= TRUE;
	DWORD				err	= ERROR_SUCCESS;
	SECURITY_ATTRIBUTES	sa	= { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
	TCHAR				lpCmdLine[1024];
	STARTUPINFO			siSlave;
	SYSTEM_INFO			siInfo;

	TRACE(trace(TEXT("*** [MASTER] OpenSlave($%08lX,%lu,%lu)\n"),lpSlave,dwInSize,dwOutSize));

	(void)memset(lpSlave,0,sizeof(*lpSlave));

	TRACE(trace(TEXT("*** [MASTER] OpenSlave(): CreateEvent()...\n")));
	if(NULL == (lpSlave->hInEvent = CreateEvent(&sa,FALSE,FALSE,NULL))) EXITONLASTERROR();
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): hInEvent = 0x%08lX\n"),lpSlave->hInEvent));

	TRACE(trace(TEXT("*** [MASTER] OpenSlave(): CreateEvent()...\n")));
	if(NULL == (lpSlave->hOutEvent = CreateEvent(&sa,FALSE,FALSE,NULL))) EXITONLASTERROR();
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): hOutEvent = 0x%08lX\n"),lpSlave->hOutEvent));

	GetSystemInfo(&siInfo);
	lpSlave->dwSize = (LPBYTE)lpSlave->lpMsg - (LPBYTE)lpSlave->lpMsg->data + dwInSize + dwOutSize;
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): dwSize = %lu\n"),lpSlave->dwSize));

	TRACE(trace(TEXT("*** [MASTER] OpenSlave(): CreateFileMapping()...\n")));
	if(NULL == (lpSlave->hFileMap = CreateFileMapping(INVALID_HANDLE_VALUE,&sa,PAGE_READWRITE,0,lpSlave->dwSize,NULL))) EXITONLASTERROR();
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): hFileMap = 0x%08lX\n"),lpSlave->hFileMap));

	TRACE(trace(TEXT("*** [MASTER] OpenSlave(): MapViewOfFile()...\n")));
	if(NULL == (lpSlave->lpMsg = (LPSLAVEMSGDATA)MapViewOfFile(lpSlave->hFileMap,FILE_MAP_WRITE,0,0,0))) EXITONLASTERROR();
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): lpMsg = 0x%08lX\n"),lpSlave->lpMsg));

	(void)memset(&siSlave,0,sizeof(siSlave)); siSlave.cb = sizeof(siSlave);

	stprintf(lpCmdLine,TEXT("0x%08lX 0x%08lX 0x%08lX"),lpSlave->hFileMap,lpSlave->hInEvent,lpSlave->hOutEvent);
	TRACE(trace(TEXT("*** [MASTER] OpenSlave(): CreateProcess(\"%s\")...\n"),lpCmdLine));
	if(!CreateProcess(TEXT("slave.exe"),lpCmdLine,NULL,NULL,TRUE,NOTHING /*CREATE_NO_WINDOW*/,NULL,NULL,&siSlave,&lpSlave->piSlave)) { (void)memset(&lpSlave->piSlave,0,sizeof(lpSlave->piSlave)); EXITONLASTERROR(); }
	TRACE(trace(TEXT(">>> [MASTER] OpenSlave(): piSlave.hProcess = 0x%08lX\n"),lpSlave->piSlave.hProcess));

exit:
	if(!z) (void)CloseSlave(lpSlave,NULL);
	TRACE(trace(TEXT("*** [MASTER] OpenSlave() >>> %s (%lu)\n"),z ? TEXT("TRUE") : TEXT("FALSE"),err));
	if(!z) SetLastError(err); return z;
}

/** @return if FALSE, use GetLastError() */
BOOL SendSlaveMsg(LPSLAVEINFO lpSlave, SLAVEMSG msg, LPVOID pvInData, DWORD dwInSize, LPVOID* ppvOutData, DWORD* pdwOutSize, HRESULT* phr)
{
	BOOL	z	= TRUE;
	DWORD	err	= ERROR_SUCCESS;

	TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg($%08lX,%d,$%08lX,%lu,...)\n"),lpSlave,pvInData,dwInSize));

	/// TODO: chek whether lpSlave is valid
	/// TODO: check whether lpSlave->piSlave.hProcess is still running

	lpSlave->lpMsg->msg			= msg;
	lpSlave->lpMsg->hr			= S_OK;
	lpSlave->lpMsg->dwInOffset	= 0;
	lpSlave->lpMsg->dwInSize	= pvInData ? dwInSize : 0;
	lpSlave->lpMsg->dwInUsed	= 0;
	lpSlave->lpMsg->dwOutOffset	= lpSlave->lpMsg->dwInOffset + lpSlave->lpMsg->dwInSize;
	if(lpSlave->lpMsg->dwOutOffset & 3)
		lpSlave->lpMsg->dwOutOffset = ((lpSlave->lpMsg->dwOutOffset >> 2) + 1) << 2;
	lpSlave->lpMsg->dwOutSize	= lpSlave->dwSize - (sizeof(*lpSlave->lpMsg) - sizeof(lpSlave->lpMsg->data)) - lpSlave->lpMsg->dwOutOffset;
	lpSlave->lpMsg->dwOutUsed	= 0;

	/// TODO: allocate the In/Out buffers in OpenSlave() so that the client
	///       can access the In buffer directly, avoiding this copy
	if(lpSlave->lpMsg->dwInSize > 0) (void)memcpy(&lpSlave->lpMsg->data[lpSlave->lpMsg->dwInOffset],pvInData,lpSlave->lpMsg->dwInSize);

	TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg(): SetEvent(hInEvent)...\n")));
	if(!SetEvent(lpSlave->hInEvent)) EXITONLASTERROR();
	TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg(): WaitForSingleObject(hOutEvent)...\n")));
	if(WAIT_FAILED == WaitForSingleObject(lpSlave->hOutEvent,INFINITE)) EXITONLASTERROR();
	TRACE(trace(TEXT(">>> [MASTER] SendSlaveMsg(): lpMsg->dwInUsed  = %lu\n"),lpSlave->lpMsg->dwInUsed));
	TRACE(trace(TEXT(">>> [MASTER] SendSlaveMsg(): lpMsg->dwOutUsed = %lu\n"),lpSlave->lpMsg->dwOutUsed));
	TRACE(trace(TEXT(">>> [MASTER] SendSlaveMsg(): lpMsg->hr ...... = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(lpSlave->lpMsg->hr)));

	if(ppvOutData)	*ppvOutData	= &lpSlave->lpMsg->data[lpSlave->lpMsg->dwOutOffset];
	if(pdwOutSize)	*pdwOutSize	= lpSlave->lpMsg->dwOutUsed;
	if(phr)			*phr		= lpSlave->lpMsg->hr;

exit:
	TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg() >>> %s (%lu)\n"),z ? TEXT("TRUE") : TEXT("FALSE"),err));
	if(!z) SetLastError(err); return z;
}


int main(int argc, char* argv[])
{
	HRESULT hr = S_OK;
	SLAVEINFO si;

	TRACE(trace_open(NULL,zTrue));

	if(!OpenSlave(&si,4096,4096)) EXITONLASTHRESULT();

	{
		LPWSTR s = L"A sample Unicode string";
		TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg(SLAVEMSG_PrintStringW,L\"%") TRACESTRW TEXT("\")...\n"),s));
		if(!SendSlaveMsg(&si,SLAVEMSG_PrintStringW,s,sizeof(*s)*(1+wcslen(s)),NULL,0,&hr)) EXITONLASTHRESULT();
		TRACE(trace(TEXT(">>> [MASTER] lpMsg->hr = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(hr)));
	}
	{
		TRACE(trace(TEXT("*** [MASTER] SendSlaveMsg(SLAVEMSG_Quit)...\n")));
		if(!SendSlaveMsg(&si,SLAVEMSG_Quit,NULL,0,NULL,0,&hr)) EXITONLASTHRESULT();
		TRACE(trace(TEXT(">>> [MASTER] lpMsg->hr = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(hr)));
	}

exit:
	TRACE(trace(TEXT(">>> [MASTER] hr = ") TRACEHR_STR TEXT("\n"),TRACEHR_REF(hr)));
	(void)CloseSlave(&si,NULL);
	TRACE(trace_close());
	return hr;
}



